"""
smbus3 - A drop-in replacement for smbus2/smbus-cffi/smbus-python
"""

import os
from ctypes import (
    POINTER,
    Structure,
    Union,
    c_char,
    c_uint8,
    c_uint16,
    c_uint32,
    create_string_buffer,
    string_at,
)
from enum import IntFlag
from fcntl import ioctl

# Commands from uapi/linux/i2c-dev.h
I2C_RETRIES = 0x0701  # Number of retries
I2C_TIMEOUT = 0x0702  # Timeout
I2C_SLAVE = 0x0703  # Use this slave address
I2C_SLAVE_FORCE = 0x0706  # Use this slave address, even if it is already in use by a driver!
I2C_TENBIT = 0x0704  # 0 for 7 bit addrs, != 0 for 10 bit
# NOTE: 10-bit address support is very limited
I2C_FUNCS = 0x0705  # Get the adapter functionality mask
I2C_RDWR = 0x0707  # Combined R/W transfer (one STOP only)
I2C_PEC = 0x0708  # != 0 to use PEC with SMBus
I2C_SMBUS = 0x0720  # SMBus transfer. Takes pointer to i2c_smbus_ioctl_data

# SMBus transfer read or write markers from uapi/linux/i2c.h
I2C_SMBUS_WRITE = 0
I2C_SMBUS_READ = 1

# Size identifiers uapi/linux/i2c.h
I2C_SMBUS_QUICK = 0
I2C_SMBUS_BYTE = 1
I2C_SMBUS_BYTE_DATA = 2
I2C_SMBUS_WORD_DATA = 3
I2C_SMBUS_PROC_CALL = 4
I2C_SMBUS_BLOCK_DATA = 5
# NOTE: I2c_SMBUS_BLOCK_DATA isn't supported by Pure-I2C drivers with
# SMBUS emulation, like those in RaspberryPi, OrangePi, etc.
I2C_SMBUS_BLOCK_PROC_CALL = 7
# Like I2C_SMBUS_BLOCK_DATA, it isn't supported by Pure-I2C drivers either.
I2C_SMBUS_I2C_BLOCK_DATA = 8
I2C_SMBUS_BLOCK_MAX = 32

# To determine what functionality is present (uapi/linux/i2c.h)


class I2cFunc(IntFlag):
    """
    These flags identify the operations supported by an I2C/SMBus device.

    You can test these flags on your `smbus.funcs`
    """

    I2C = 0x00000001
    ADDR_10BIT = 0x00000002
    PROTOCOL_MANGLING = 0x00000004  # I2C_M_IGNORE_NAK etc.
    SMBUS_PEC = 0x00000008
    NOSTART = 0x00000010  # I2C_M_NOSTART
    SLAVE = 0x00000020
    SMBUS_BLOCK_PROC_CALL = 0x00008000  # SMBus 2.0
    SMBUS_QUICK = 0x00010000
    SMBUS_READ_BYTE = 0x00020000
    SMBUS_WRITE_BYTE = 0x00040000
    SMBUS_READ_BYTE_DATA = 0x00080000
    SMBUS_WRITE_BYTE_DATA = 0x00100000
    SMBUS_READ_WORD_DATA = 0x00200000
    SMBUS_WRITE_WORD_DATA = 0x00400000
    SMBUS_PROC_CALL = 0x00800000
    SMBUS_READ_BLOCK_DATA = 0x01000000
    SMBUS_WRITE_BLOCK_DATA = 0x02000000
    SMBUS_READ_I2C_BLOCK = 0x04000000  # I2C-like block xfer
    SMBUS_WRITE_I2C_BLOCK = 0x08000000  # w/ 1-byte reg. addr.
    SMBUS_HOST_NOTIFY = 0x10000000

    SMBUS_BYTE = 0x00060000
    SMBUS_BYTE_DATA = 0x00180000
    SMBUS_WORD_DATA = 0x00600000
    SMBUS_BLOCK_DATA = 0x03000000
    SMBUS_I2C_BLOCK = 0x0C000000
    SMBUS_EMUL = 0x0EFF0008


# i2c_msg flags from uapi/linux/i2c.h
I2C_M_RD = 0x0001
I2C_M_WR = 0x0000
I2C_M_TEN = 0x0010


class I2C_M_Bitflag(IntFlag):
    """
    These flags identify the operations to perform with
    a specific i2c_msg.
    """

    I2C_M_RD = 0x0001
    I2C_M_WR = 0x0000
    I2C_M_TEN = 0x0010
    I2C_M_RD_TEN = I2C_M_RD ^ I2C_M_TEN
    I2C_M_WR_TEN = I2C_M_WR ^ I2C_M_TEN


# Pointer definitions
LP_c_uint8 = POINTER(c_uint8)
LP_c_uint16 = POINTER(c_uint16)
LP_c_uint32 = POINTER(c_uint32)


i2c_smbus_data = c_uint8 * (I2C_SMBUS_BLOCK_MAX + 2)
"""
Adaptation of the i2c_smbus_data union in ``i2c.h``.
Data for SMBus messages.

Add two additional blocks, one for length, one for user-space
compatibility.
"""


class union_i2c_smbus_data(Union):
    """
    C union for representing i2c smbus data.
    """

    _fields_ = [("byte", c_uint8), ("word", c_uint16), ("block", i2c_smbus_data)]

    def __repr__(self):
        return (
            "<smbus3.smbus3.union_i2c_smbus_data "
            f"byte: {self.byte} word: {self.word} block: {self.block}>"
        )


union_pointer_type = POINTER(union_i2c_smbus_data)


class i2c_smbus_ioctl_data(Structure):
    """
    As defined in ``i2c-dev.h``.
    """

    _fields_ = [
        ("read_write", c_uint8),
        ("command", c_uint8),
        ("size", c_uint32),
        ("data", union_pointer_type),
    ]
    __slots__ = [name for name, type in _fields_]

    def __repr__(self):
        return (
            "<smbus3.smbus3.i2c_smbus_ioctl_data "
            f"read_write: {self.read_write} command: {self.command} "
            f"size: {self.size} data: {self.data.contents!r}>"
        )

    @staticmethod
    def create(read_write=I2C_SMBUS_READ, command=0, size=I2C_SMBUS_BYTE_DATA):
        """
        Create a new i2c_smbus_ioctl_data struct.
        """
        u = union_i2c_smbus_data()
        return i2c_smbus_ioctl_data(
            read_write=read_write,
            command=command,
            size=size,
            data=union_pointer_type(u),
        )


class i2c_msg(Structure):
    """
    As defined in ``i2c.h``.

    :ivar addr: i2c address
    :vartype addr: c_uint16
    :ivar flags: bit flags
    :vartype flags: c_uint16
    :ivar len: length
    :vartype len: c_uint16
    :ivar buf: data buffer
    :vartype buf: POINTER(c_char)
    """

    _fields_ = [
        ("addr", c_uint16),
        ("flags", c_uint16),
        ("len", c_uint16),
        ("buf", POINTER(c_char)),
    ]

    def __iter__(self):
        """
        Iterator / Generator

        :return: iterates over :py:attr:`buf`
        :rtype: :py:class:`generator` which returns int values
        """
        idx = 0
        while idx < self.len:
            yield ord(self.buf[idx])
            idx += 1

    def __len__(self):
        return self.len

    def __bytes__(self):
        return string_at(self.buf, self.len)

    def __repr__(self):
        return f"i2c_msg({self.addr:d},{self.flags:x},{self.__bytes__()!r})"

    def __str__(self):
        """
        Return a string, discarding non decodable bytes.

        :return: string of i2c_msg.__bytes__()
        :rtype: str
        """
        s = self.__bytes__()
        # Throw away non-decodable bytes
        s = s.decode(errors="ignore")
        return s

    @staticmethod
    def read(address, length, flags=I2C_M_RD):
        """
        Prepares an i2c read transaction.

        :param address: Slave address.
        :type: address: int
        :param length: Number of bytes to read.
        :type: length: int
        :param flags: bitflags to pass (default: I2C_M_RD)
        :type flags: int
        :return: New :py:class:`i2c_msg` instance for read operation.
        :rtype: :py:class:`i2c_msg`
        """
        arr = create_string_buffer(length)
        return i2c_msg(addr=address, flags=flags, len=length, buf=arr)

    @staticmethod
    def write(address, buf, flags=I2C_M_WR):
        """
        Prepares an i2c write transaction.

        :param address: Slave address.
        :type address: int
        :param buf: Bytes to write. Either list of values or str.
        :type buf: list
        :param flags: bitflags to pass (default: I2C_M_WR)
        :type flags: int
        :return: New :py:class:`i2c_msg` instance for write operation.
        :rtype: :py:class:`i2c_msg`
        """
        if isinstance(buf, str):
            buf = bytes(map(ord, buf))
        else:
            buf = bytes(buf)
        arr = create_string_buffer(buf, len(buf))
        return i2c_msg(addr=address, flags=flags, len=len(arr), buf=arr)


class i2c_rdwr_ioctl_data(Structure):
    """
    As defined in ``i2c-dev.h``.

    Fields:
        msgs  - POINTER(i2c_msg)

        nmsgs - c_uint32
    """

    _fields_ = [("msgs", POINTER(i2c_msg)), ("nmsgs", c_uint32)]
    __slots__ = [name for name, type in _fields_]

    @staticmethod
    def create(*i2c_msg_instances):
        """
        Factory method for creating a i2c_rdwr_ioctl_data struct that can
        be called with ``ioctl(fd, I2C_RDWR, data)``.

        :param i2c_msg_instances: Up to 42 i2c_msg instances
        :rtype: i2c_rdwr_ioctl_data
        """
        n_msg = len(i2c_msg_instances)
        msg_array = (i2c_msg * n_msg)(*i2c_msg_instances)
        return i2c_rdwr_ioctl_data(msgs=msg_array, nmsgs=n_msg)


class SMBus:
    """
    The main SMBus class.
    """

    def __init__(self, bus=None, force=False):
        """
        Initialize and (optionally) open an i2c bus connection.

        :param bus: i2c bus number (e.g. 0 or 1)
            or an absolute file path (e.g. `/dev/i2c-42`).
            If not given, a subsequent  call to ``open()`` is required.
        :type bus: int or str
        :param force: force using the slave address even when driver is
            already using it.
        :type force: boolean
        """
        self.fd = None
        self.funcs = I2cFunc(0)
        if bus is not None:
            self.open(bus)
        self.address = None
        self.force = force
        self._force_last = None
        self._pec = 0
        self._tenbit = 0

    def __enter__(self):
        """Enter handler."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit handler."""
        self.close()

    def open(self, bus):
        """
        Open a given i2c bus.

        :param bus: i2c bus number (e.g. 0 or 1)
            or an absolute file path (e.g. '/dev/i2c-42').
        :type bus: int or str
        :raise TypeError: if type(bus) is not in (int, str)
        :rtype: None
        """
        if isinstance(bus, int):
            filepath = f"/dev/i2c-{bus}"
        elif isinstance(bus, str):
            filepath = bus
        else:
            raise TypeError(f"Unexpected type(bus)={type(bus)}")

        self.fd = os.open(filepath, os.O_RDWR)
        self.funcs = self._get_funcs()

    def close(self):
        """
        Close the i2c connection.

        :raise OSError: if the file descriptor in self.fd does not exist
        :rtype: None
        """
        if self.fd:
            os.close(self.fd)
            self.fd = None
            self._pec = 0
            self._tenbit = 0
            self.address = None
            self._force_last = None

    def _get_pec(self):
        return self._pec

    def enable_pec(self, enable=True):
        """
        Enable/Disable PEC (Packet Error Checking) - SMBus 1.1 and later

        :param enable: Whether to enable PEC or not.
        :type enable: bool
        :raise IOError: if SMBUS_PEC is not supported.
        :rtype: None
        """
        if not (self.funcs & I2cFunc.SMBUS_PEC):
            raise OSError("SMBUS_PEC is not a feature")
        self._pec = int(enable)
        ioctl(self.fd, I2C_PEC, self._pec)

    pec = property(_get_pec, enable_pec)  # Drop-in replacement for smbus member "pec"
    """Get and set SMBus PEC. 0 = disabled (default), 1 = enabled."""

    def _get_tenbit(self):
        return self._tenbit

    def enable_tenbit(self, enable=True):
        """
        Enable 10 bit addresses if they are supported.

        :param enable: Whether to enable 10bit addresses or not.
        :type enable: bool
        :raise IOError: if ADDR_10BIT is not supported.
        :rtype: None
        """
        if not (self.funcs & I2cFunc.ADDR_10BIT):
            raise OSError("ADDR_10BIT is not a feature")
        self._tenbit = int(enable)
        ioctl(self.fd, I2C_TENBIT, self._tenbit)

    tenbit = property(_get_tenbit, enable_tenbit)
    """Get and set 10bit addressing. 0 = disabled (default), 1 = enabled."""

    def _get_timeout(self):
        return self._timeout

    def set_timeout(self, timeout):
        """
        Set the timeout in units of 10ms.

        :param timeout: timeout as positive int in units of 10ms.
        :type timeout: int
        :rtype: None
        """
        self._timeout = timeout
        ioctl(self.fd, I2C_TIMEOUT, self._timeout)

    timeout = property(_get_timeout, set_timeout)
    """Get and set I2C timeout in units of 10ms."""

    def _get_retries(self):
        return self._retries

    def set_retries(self, retries):
        """
        Set the retries.

        :param retries: retries as positive int.
        :type retries: int
        :rtype: None
        """
        self._retries = retries
        ioctl(self.fd, I2C_RETRIES, self._retries)

    retries = property(_get_retries, set_retries)
    """Get and set I2C retries."""

    def _set_address(self, address, force=None):
        """
        Set i2c slave address to use for subsequent calls.
        Private.

        :param address:
        :type address: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :rtype: None
        """
        force = force if force is not None else self.force
        if self.address != address or self._force_last != force:
            if force is True:
                ioctl(self.fd, I2C_SLAVE_FORCE, address)
            else:
                ioctl(self.fd, I2C_SLAVE, address)
            self.address = address
            self._force_last = force

    def _get_funcs(self):
        """
        Returns a 32-bit value stating supported I2C functions.
        Private.

        :rtype: int
        """
        f = c_uint32()
        ioctl(self.fd, I2C_FUNCS, f)
        return f.value

    def write_quick(self, i2c_addr, force=None):
        """
        Perform quick transaction. Throws IOError if unsuccessful.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :raise IOError: if write is unsuccessful.
        :rtype: None
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_WRITE, command=0, size=I2C_SMBUS_QUICK
        )
        ioctl(self.fd, I2C_SMBUS, msg)

    def read_byte(self, i2c_addr, force=None):
        """
        Read a single byte from a device.

        :rtype: int
        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :return: Read byte value
        :rtype: int
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(read_write=I2C_SMBUS_READ, command=0, size=I2C_SMBUS_BYTE)
        ioctl(self.fd, I2C_SMBUS, msg)
        return msg.data.contents.byte

    def write_byte(self, i2c_addr, value, force=None):
        """
        Write a single byte to a device.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param value: value to write
        :type value: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :rtype: None
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_WRITE, command=value, size=I2C_SMBUS_BYTE
        )
        ioctl(self.fd, I2C_SMBUS, msg)

    def read_byte_data(self, i2c_addr, register, force=None):
        """
        Read a single byte from a designated register.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Register to read
        :type register: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :return: Read byte value
        :rtype: int
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_READ, command=register, size=I2C_SMBUS_BYTE_DATA
        )
        ioctl(self.fd, I2C_SMBUS, msg)
        return msg.data.contents.byte

    def write_byte_data(self, i2c_addr, register, value, force=None):
        """
        Write a byte to a given register.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Register to write to
        :type register: int
        :param value: Byte value to transmit
        :type value: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :rtype: None
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_WRITE, command=register, size=I2C_SMBUS_BYTE_DATA
        )
        msg.data.contents.byte = value
        ioctl(self.fd, I2C_SMBUS, msg)

    def read_word_data(self, i2c_addr, register, force=None):
        """
        Read a single word (2 bytes) from a given register.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Register to read
        :type register: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :return: 2-byte word
        :rtype: int
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_READ, command=register, size=I2C_SMBUS_WORD_DATA
        )
        ioctl(self.fd, I2C_SMBUS, msg)
        return msg.data.contents.word

    def write_word_data(self, i2c_addr, register, value, force=None):
        """
        Write a single word (2 bytes) to a given register.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Register to write to
        :type register: int
        :param value: Word value to transmit
        :type value: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :rtype: None
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_WRITE, command=register, size=I2C_SMBUS_WORD_DATA
        )
        msg.data.contents.word = value
        ioctl(self.fd, I2C_SMBUS, msg)

    def process_call(self, i2c_addr, register, value, force=None):
        """
        Executes a SMBus Process Call, sending a 16-bit value and receiving a 16-bit response

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Register to read/write to
        :type register: int
        :param value: Word value to transmit
        :type value: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :rtype: int
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_WRITE, command=register, size=I2C_SMBUS_PROC_CALL
        )
        msg.data.contents.word = value
        ioctl(self.fd, I2C_SMBUS, msg)
        return msg.data.contents.word

    def read_block_data(self, i2c_addr, register, force=None):
        """
        Read a block of up to 32-bytes from a given register.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Start register
        :type register: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :return: List of bytes
        :rtype: list
        """
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_READ, command=register, size=I2C_SMBUS_BLOCK_DATA
        )
        ioctl(self.fd, I2C_SMBUS, msg)
        length = msg.data.contents.block[0]
        return msg.data.contents.block[1 : length + 1]

    def write_block_data(self, i2c_addr, register, data, force=None):
        """
        Write a block of byte data to a given register.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Start register
        :type register: int
        :param data: List of bytes
        :type data: list
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :raise ValueError: if length of data in bytes is > I2C_SMBUS_BLOCK_MAX
        :rtype: None
        """
        length = len(data)
        if length > I2C_SMBUS_BLOCK_MAX:
            raise ValueError(f"Data length cannot exceed {I2C_SMBUS_BLOCK_MAX:d} bytes")
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_WRITE, command=register, size=I2C_SMBUS_BLOCK_DATA
        )
        msg.data.contents.block[0] = length
        msg.data.contents.block[1 : length + 1] = data
        ioctl(self.fd, I2C_SMBUS, msg)

    def block_process_call(self, i2c_addr, register, data, force=None):
        """
        Executes a SMBus Block Process Call, sending a variable-size data
        block and receiving another variable-size response

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Register to read/write to
        :type register: int
        :param data: List of bytes
        :type data: list
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :raise ValueError: if length of data in bytes is > I2C_SMBUS_BLOCK_MAX
        :return: List of bytes
        :rtype: list
        """
        length = len(data)
        if length > I2C_SMBUS_BLOCK_MAX:
            raise ValueError(f"Data length cannot exceed {I2C_SMBUS_BLOCK_MAX:d} bytes")
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_WRITE, command=register, size=I2C_SMBUS_BLOCK_PROC_CALL
        )
        msg.data.contents.block[0] = length
        msg.data.contents.block[1 : length + 1] = data
        ioctl(self.fd, I2C_SMBUS, msg)
        length = msg.data.contents.block[0]
        return msg.data.contents.block[1 : length + 1]

    def read_i2c_block_data(self, i2c_addr, register, length, force=None):
        """
        Read a block of byte data from a given register.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Start register
        :type register: int
        :param length: Desired block length
        :type length: int
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :raise ValueError: if length (in bytes) is > I2C_SMBUS_BLOCK_MAX
        :return: List of bytes
        :rtype: list
        """
        if length > I2C_SMBUS_BLOCK_MAX:
            raise ValueError(f"Data length cannot exceed {I2C_SMBUS_BLOCK_MAX:d} bytes")
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_READ, command=register, size=I2C_SMBUS_I2C_BLOCK_DATA
        )
        msg.data.contents.byte = length
        ioctl(self.fd, I2C_SMBUS, msg)
        return msg.data.contents.block[1 : length + 1]

    def write_i2c_block_data(self, i2c_addr, register, data, force=None):
        """
        Write a block of byte data to a given register.

        :param i2c_addr: i2c address
        :type i2c_addr: int
        :param register: Start register
        :type register: int
        :param data: List of bytes
        :type data: list
        :param force: force using the slave address even when driver is already using it.
        :type force: bool
        :raise ValueError: if length of data in bytes is > I2C_SMBUS_BLOCK_MAX
        :rtype: None
        """
        length = len(data)
        if length > I2C_SMBUS_BLOCK_MAX:
            raise ValueError(f"Data length cannot exceed {I2C_SMBUS_BLOCK_MAX:d} bytes")
        self._set_address(i2c_addr, force=force)
        msg = i2c_smbus_ioctl_data.create(
            read_write=I2C_SMBUS_WRITE, command=register, size=I2C_SMBUS_I2C_BLOCK_DATA
        )
        msg.data.contents.block[0] = length
        msg.data.contents.block[1 : length + 1] = data
        ioctl(self.fd, I2C_SMBUS, msg)

    def i2c_rdwr(self, *i2c_msgs):
        """
        Combine a series of i2c read and write operations in a single
        transaction (with repeated start bits but no stop bits in between).

        This method takes i2c_msg instances as input, which must be created
        first with :py:meth:`i2c_msg.read` or :py:meth:`i2c_msg.write`.

        :param i2c_msgs: One or more i2c_msg class instances.
        :type i2c_msgs: i2c_msg
        :rtype: None
        """
        ioctl_data = i2c_rdwr_ioctl_data.create(*i2c_msgs)
        ioctl(self.fd, I2C_RDWR, ioctl_data)

    def i2c_rd(self, i2c_addr, length, flags=I2C_M_RD):
        """
        Perform a single i2c read operation, given an i2c_addr and length.

        :param i2c_addr: i2c address for the read operation.
        :type i2c_addr: int
        :param length: length of read.
        :type length: int
        :param flags: bitflags to pass (default: I2C_M_RD)
        :type flags: int
        :rtype: i2c_msg
        """
        msg = i2c_msg.read(i2c_addr, length, flags=flags)
        self.i2c_rdwr(msg)
        return msg

    def i2c_wr(self, i2c_addr, buf, flags=I2C_M_WR):
        """
        Perform a single i2c write operation, given an i2c_addr and a
        buffer to copy.

        :param i2c_addr: i2c address for the write operation.
        :type i2c_addr: int
        :param buf: buffer to write.
        :type buf: list
        :param flags: bitflags to pass (default: I2C_M_WR)
        :type flags: int
        :rtype: i2c_msg
        """
        msg = i2c_msg.write(i2c_addr, buf, flags=flags)
        self.i2c_rdwr(msg)
        return msg



"""
smbus3 - A drop-in replacement for smbus2/smbus-cffi/smbus-python
"""

from .smbus3 import I2C_M_Bitflag, I2cFunc, SMBus, i2c_msg

__version__ = "0.5.4"
__all__ = ["SMBus", "i2c_msg", "I2cFunc", "I2C_M_Bitflag", "__version__"]



"""
tests/test_smbus3.py
--------------------

Main tests for SMBus class, i2c_msg, and I2cFunc.
"""

import unittest
from contextlib import contextmanager
from unittest import mock

from smbus3 import I2C_M_Bitflag, I2cFunc, SMBus, i2c_msg

# Required I2C constant definitions repeated
I2C_FUNCS = 0x0705  # Get the adapter functionality mask
I2C_SMBUS = 0x0720
I2C_SMBUS_WRITE = 0
I2C_SMBUS_READ = 1

I2C_SMBUS_QUICK = 0
I2C_SMBUS_BYTE = 1
I2C_SMBUS_BYTE_DATA = 2
I2C_SMBUS_WORD_DATA = 3
I2C_SMBUS_BLOCK_DATA = 5
I2C_SMBUS_I2C_BLOCK_DATA = 8
I2C_SMBUS_BLOCK_MAX = 32

MOCK_FD = "Mock file descriptor"
MOCK_I2C_FUNC_LIMITED = 0xEFF0001
MOCK_I2C_FUNC_FULL = 0xEFF000B
MOCK_MSG = None

# Test buffer for read operations
test_buffer = [x for x in range(256)]


def mock_msg_refresh(msg):
    """
    Cleanup the mock msg used for writes.
    """
    global MOCK_MSG  # noqa
    MOCK_MSG = msg


# Mock open, close and ioctl so we can run our unit tests anywhere.
def mock_open(*args):
    print(f"Mocking open: {args[0]}")
    return MOCK_FD


def mock_close(*args):
    print("Mocking close")
    assert args[0] == MOCK_FD


def mock_ioctl_limited(fd, command, msg):
    print(f"Mocking ioctl with command 0x{command:X} and msg {msg}")
    assert fd == MOCK_FD
    assert command is not None
    assert isinstance(command, int)

    # Reproduce i2c capability of a Raspberry Pi 3 w/o PEC support and w/o
    # 10bit addressing support
    if command == I2C_FUNCS:
        msg.value = MOCK_I2C_FUNC_LIMITED
        print(f"Setting msg val: 0x{msg.value:X}")
        return

    # Reproduce ioctl read operations
    if command == I2C_SMBUS and msg.read_write == I2C_SMBUS_READ:
        offset = msg.command
        if msg.size == I2C_SMBUS_BYTE:
            msg.data.contents.byte = test_buffer[offset]
        elif msg.size == I2C_SMBUS_BYTE_DATA:
            msg.data.contents.byte = test_buffer[offset]
        elif msg.size == I2C_SMBUS_WORD_DATA:
            msg.data.contents.word = test_buffer[offset + 1] * 256 + test_buffer[offset]
        elif msg.size == I2C_SMBUS_I2C_BLOCK_DATA:
            for k in range(msg.data.contents.byte):
                msg.data.contents.block[k + 1] = test_buffer[offset + k]
        elif msg.size == I2C_SMBUS_BLOCK_DATA:
            msg.data.contents.block[0] = 32
            for k in range(32):
                msg.data.contents.block[k + 1] = test_buffer[offset + k]

    # Reproduce ioctl write operations
    if command == I2C_SMBUS and msg.read_write == I2C_SMBUS_WRITE:
        mock_msg_refresh(msg)

    # Reproduce a failing Quick write transaction
    if command == I2C_SMBUS and msg.read_write == I2C_SMBUS_WRITE and msg.size == I2C_SMBUS_QUICK:
        raise OSError("Mocking SMBus Quick failed")


def mock_ioctl_full(fd, command, msg):
    print(f"Mocking ioctl with command 0x{command:X} and msg {msg}")
    assert fd == MOCK_FD
    assert command is not None
    assert isinstance(command, int)

    # Reproduce i2c capability w/ PEC support and w/
    # 10bit addressing support
    if command == I2C_FUNCS:
        msg.value = MOCK_I2C_FUNC_FULL
        print(f"Setting msg val: 0x{msg.value:X}")
        return

    # Reproduce ioctl read operations
    if command == I2C_SMBUS and msg.read_write == I2C_SMBUS_READ:
        offset = msg.command
        if msg.size == I2C_SMBUS_BYTE_DATA:
            msg.data.contents.byte = test_buffer[offset]
        elif msg.size == I2C_SMBUS_WORD_DATA:
            msg.data.contents.word = test_buffer[offset + 1] * 256 + test_buffer[offset]
        elif msg.size == I2C_SMBUS_I2C_BLOCK_DATA:
            for k in range(msg.data.contents.byte):
                msg.data.contents.block[k + 1] = test_buffer[offset + k]
        elif msg.size == I2C_SMBUS_BLOCK_DATA:
            msg.data.contents.block[0] = 32
            for k in range(32):
                msg.data.contents.block[k + 1] = test_buffer[offset + k]

    # Reproduce ioctl write operations
    if command == I2C_SMBUS and msg.read_write == I2C_SMBUS_WRITE:
        mock_msg_refresh(msg)


# Override open, close, read, and ioctl with our mock functions
open_mock = mock.patch("smbus3.smbus3.os.open", mock_open)
close_mock = mock.patch("smbus3.smbus3.os.close", mock_close)
ioctl_limited_mock = mock.patch("smbus3.smbus3.ioctl", mock_ioctl_limited)
ioctl_full_mock = mock.patch("smbus3.smbus3.ioctl", mock_ioctl_full)


@contextmanager
def switch_to_full_featured_ioctl_mock():
    """
    This context manager allows you to temporarily switch the I2C_FUNCS supported
    by the ioctl mock used in the test.
    """
    try:
        ioctl_limited_mock.stop()
        ioctl_full_mock.start()
        yield
    finally:
        ioctl_full_mock.stop()
        ioctl_limited_mock.start()


# Common error messages
INCORRECT_LENGTH_MSG = "Result array of incorrect length."


class SMBusTestCase(unittest.TestCase):
    def setUp(self):
        open_mock.start()
        close_mock.start()
        ioctl_limited_mock.start()

    def tearDown(self):
        open_mock.stop()
        close_mock.stop()
        ioctl_limited_mock.stop()


# Test cases
class TestSMBus(SMBusTestCase):
    def test_func(self):
        bus = SMBus(1)
        print(f"Supported I2C functionality: 0x{bus.funcs:X}")
        self.assertEqual(bus.funcs, MOCK_I2C_FUNC_LIMITED)
        bus.close()

        # Now try with full features:
        with switch_to_full_featured_ioctl_mock():
            bus = SMBus(1)
            print(f"Supported I2C functionality: 0x{bus.funcs:X}")
            self.assertEqual(bus.funcs, MOCK_I2C_FUNC_FULL)
            bus.close()

    def test_enter_exit(self):
        for idx in (1, "/dev/i2c-alias"):
            with SMBus(idx) as bus:
                self.assertIsNotNone(bus.fd)
            self.assertIsNone(bus.fd, None)

        with SMBus() as bus:
            self.assertIsNone(bus.fd)
            bus.open(2)
            self.assertIsNotNone(bus.fd)
        self.assertIsNone(bus.fd)

    def test_open_close(self):
        for idx in (1, "/dev/i2c-alias"):
            bus = SMBus()
            self.assertIsNone(bus.fd)
            bus.open(idx)
            self.assertIsNotNone(bus.fd)
            bus.close()
            self.assertIsNone(bus.fd)
        bus = SMBus()
        with self.assertRaises(TypeError):
            bus.open([1, 2])

    def test_write(self):  # noqa: PLR0915
        bus = SMBus(1)
        # Write byte:
        bus.write_byte(80, 0x001, force=False)
        self.assertEqual(MOCK_MSG.read_write, 0)
        self.assertEqual(MOCK_MSG.command, 1)
        self.assertEqual(MOCK_MSG.size, 1)
        mock_msg_refresh(None)
        # Write byte - force = True
        bus.write_byte(80, 0x001, force=True)
        self.assertEqual(MOCK_MSG.read_write, 0)
        self.assertEqual(MOCK_MSG.command, 1)
        self.assertEqual(MOCK_MSG.size, 1)
        mock_msg_refresh(None)
        # Write byte data
        bus.write_byte_data(80, 1, 0x001)
        self.assertEqual(MOCK_MSG.read_write, 0)
        self.assertEqual(MOCK_MSG.command, 1)
        self.assertEqual(MOCK_MSG.size, 2)
        self.assertEqual(MOCK_MSG.data.contents.byte, 1)
        self.assertEqual(MOCK_MSG.data.contents.word, 1)
        mock_msg_refresh(None)
        # Write word data:
        bus.write_word_data(80, 1, 0x001, force=False)
        self.assertEqual(MOCK_MSG.read_write, 0)
        self.assertEqual(MOCK_MSG.command, 1)
        self.assertEqual(MOCK_MSG.size, 3)
        self.assertEqual(MOCK_MSG.data.contents.byte, 1)
        self.assertEqual(MOCK_MSG.data.contents.word, 1)
        mock_msg_refresh(None)
        # Process call:
        self.assertEqual(bus.process_call(80, 1, 0x001), 1)
        self.assertEqual(MOCK_MSG.read_write, 0)
        self.assertEqual(MOCK_MSG.command, 1)
        self.assertEqual(MOCK_MSG.size, 4)
        self.assertEqual(MOCK_MSG.data.contents.byte, 1)
        self.assertEqual(MOCK_MSG.data.contents.word, 1)
        mock_msg_refresh(None)
        # Write block data:
        bus.write_block_data(80, 1, [1, 2, 3])
        self.assertEqual(MOCK_MSG.read_write, 0)
        self.assertEqual(MOCK_MSG.command, 1)
        self.assertEqual(MOCK_MSG.size, 5)
        self.assertEqual(MOCK_MSG.data.contents.byte, 3)
        self.assertEqual(MOCK_MSG.data.contents.word, 256 + 3)
        self.assertListEqual(list(MOCK_MSG.data.contents.block[1:4]), [1, 2, 3])
        mock_msg_refresh(None)
        with self.assertRaises(ValueError):
            # Try writing too large of a block
            x = [_ for _ in range(35)]
            bus.write_block_data(80, 1, x)
        # Write i2c block data:
        bus.write_i2c_block_data(80, 1, [1, 2, 3])
        self.assertEqual(MOCK_MSG.read_write, 0)
        self.assertEqual(MOCK_MSG.command, 1)
        self.assertEqual(MOCK_MSG.size, 8)
        self.assertEqual(MOCK_MSG.data.contents.byte, 3)
        self.assertEqual(MOCK_MSG.data.contents.word, 256 + 3)
        self.assertListEqual(list(MOCK_MSG.data.contents.block[1:4]), [1, 2, 3])
        mock_msg_refresh(None)
        with self.assertRaises(ValueError):
            # Try writing too large of a block
            x = [_ for _ in range(35)]
            bus.write_i2c_block_data(80, 1, x)
        # Block process call:
        x = [_ for _ in range(8)]
        self.assertEqual(bus.block_process_call(80, 1, x), x)
        with self.assertRaises(ValueError):
            # Try writing too large of a block
            x = [_ for _ in range(35)]
            bus.block_process_call(80, 1, x)
        bus.close()

    def test_read(self):
        res = []
        res2 = []
        res3 = []
        res4 = []
        res5 = []
        res6 = []

        bus = SMBus(1)

        # Read bytes
        for k in range(2):
            x = bus.read_byte_data(80, k)
            res.append(x)
        self.assertEqual(len(res), 2, msg=INCORRECT_LENGTH_MSG)

        # Read word
        x = bus.read_word_data(80, 0)
        res2.append(x & 255)
        res2.append(x / 256)
        self.assertEqual(len(res2), 2, msg=INCORRECT_LENGTH_MSG)
        self.assertListEqual(res, res2, msg="Byte and word reads differ")

        # Read block of N bytes (i2c)
        n = 2
        x = bus.read_i2c_block_data(80, 0, n)
        res3.extend(x)
        self.assertEqual(len(res3), n, msg=INCORRECT_LENGTH_MSG)
        self.assertListEqual(res, res3, msg="Byte and block reads differ")

        # Check that reading too many bytes causes a ValueError:
        with self.assertRaises(ValueError):
            bus.read_i2c_block_data(80, 0, 35)

        # Read byte - force False
        for _ in range(2):
            x = bus.read_byte(80, force=False)
            res4.append(x)
        self.assertEqual(len(res4), 2, msg=INCORRECT_LENGTH_MSG)

        # Read byte - force True
        for _ in range(2):
            x = bus.read_byte(80, force=True)
            res5.append(x)
        self.assertEqual(len(res5), 2, msg=INCORRECT_LENGTH_MSG)

        # Read block of 32 bytes
        x = bus.read_block_data(80, 0)
        res6.extend(x)
        self.assertEqual(len(res6), 32, msg=INCORRECT_LENGTH_MSG)
        self.assertListEqual(res, res6[:2], msg="Byte and block reads differ")
        self.assertEqual(
            sum(res6), sum(_ for _ in range(32)), msg="Incorrect values from read_block_data"
        )

        bus.close()

    def test_write_full(self):
        """
        Test writes with 10bit + PEC enabled.
        """
        with switch_to_full_featured_ioctl_mock():
            bus = SMBus(1)
            bus.pec = 1
            bus.tenbit = 1
            # Write byte:
            bus.write_byte(80, 0x001, force=False)
            # Write byte - force = True
            bus.write_byte(80, 0x001, force=True)
            # Write byte data
            bus.write_byte_data(80, 1, 0x001)
            # Write word data:
            bus.write_word_data(80, 1, 0x001, force=False)
            # Process call:
            self.assertEqual(bus.process_call(80, 1, 0x001), 1)
            # Write block data:
            bus.write_block_data(80, 1, [1, 2, 3])
            with self.assertRaises(ValueError):
                # Try writing too large of a block
                x = [_ for _ in range(35)]
                bus.write_block_data(80, 1, x)
            # Write i2c block data:
            bus.write_i2c_block_data(80, 1, [1, 2, 3])
            with self.assertRaises(ValueError):
                # Try writing too large of a block
                x = [_ for _ in range(35)]
                bus.write_i2c_block_data(80, 1, x)
            # Block process call:
            x = [_ for _ in range(8)]
            self.assertEqual(bus.block_process_call(80, 1, x), x)
            with self.assertRaises(ValueError):
                # Try writing too large of a block
                x = [_ for _ in range(35)]
                bus.block_process_call(80, 1, x)
            bus.close()

    def test_read_full(self):
        """
        Test reads with 10bit + PEC enabled.
        """
        with switch_to_full_featured_ioctl_mock():
            res = []
            res2 = []
            res3 = []
            res4 = []
            res5 = []
            res6 = []

            bus = SMBus(1)
            bus.pec = 1
            bus.tenbit = 1

            # Read bytes
            for k in range(2):
                x = bus.read_byte_data(80, k)
                res.append(x)
            self.assertEqual(len(res), 2, msg=INCORRECT_LENGTH_MSG)

            # Read word
            x = bus.read_word_data(80, 0)
            res2.append(x & 255)
            res2.append(x / 256)
            self.assertEqual(len(res2), 2, msg=INCORRECT_LENGTH_MSG)
            self.assertListEqual(res, res2, msg="Byte and word reads differ")

            # Read block of N bytes (i2c)
            n = 2
            x = bus.read_i2c_block_data(80, 0, n)
            res3.extend(x)
            self.assertEqual(len(res3), n, msg=INCORRECT_LENGTH_MSG)
            self.assertListEqual(res, res3, msg="Byte and block reads differ")

            # Check that reading too many bytes causes a ValueError:
            with self.assertRaises(ValueError):
                bus.read_i2c_block_data(80, 0, 35)

            # Read byte - force False
            for _ in range(2):
                x = bus.read_byte(80, force=False)
                res4.append(x)
            self.assertEqual(len(res4), 2, msg=INCORRECT_LENGTH_MSG)

            # Read byte - force True
            for _ in range(2):
                x = bus.read_byte(80, force=True)
                res5.append(x)
            self.assertEqual(len(res5), 2, msg=INCORRECT_LENGTH_MSG)

            # Read block of 32 bytes
            x = bus.read_block_data(80, 0)
            res6.extend(x)
            self.assertEqual(len(res6), 32, msg=INCORRECT_LENGTH_MSG)
            self.assertListEqual(res, res6[:2], msg="Byte and block reads differ")
            self.assertEqual(
                sum(res6), sum(_ for _ in range(32)), msg="Incorrect values from read_block_data"
            )

            bus.close()

    def test_quick(self):
        bus = SMBus(1)
        self.assertRaises(IOError, bus.write_quick, 80)

    def test_timeout(self):
        def set_timeout(bus, timeout=15):
            bus.timeout = timeout

        bus = SMBus(1)
        set_timeout(bus)

        self.assertEqual(bus._timeout, 15)
        self.assertEqual(bus._get_timeout(), 15)
        self.assertEqual(bus.timeout, 15)
        bus.close()

    def test_retries(self):
        def set_retries(bus, retries=3):
            bus.retries = retries

        bus = SMBus(1)
        set_retries(bus)

        self.assertEqual(bus._retries, 3)
        self.assertEqual(bus._get_retries(), 3)
        self.assertEqual(bus.retries, 3)
        bus.close()

    def test_pec(self):
        def set_pec(bus, enable=True):
            bus.pec = enable

        # Enabling PEC should fail (no mocked PEC support)
        bus = SMBus(1)
        self.assertRaises(IOError, set_pec, bus, True)
        self.assertRaises(IOError, set_pec, bus, 1)
        self.assertEqual(bus.pec, 0)

        # Ensure PEC status is reset by close()
        bus._pec = 1
        self.assertEqual(bus.pec, 1)
        bus.close()
        self.assertEqual(bus.pec, 0)

        # Test true enable by changing ioctl mock:
        with switch_to_full_featured_ioctl_mock():
            bus = SMBus(1)
            self.assertEqual(bus.pec, 0)
            set_pec(bus, enable=True)
            self.assertEqual(bus.pec, 1)
        bus.close()
        self.assertEqual(bus.pec, 0)

    def test_tenbit(self):
        def set_tenbit(bus, enable=True):
            bus.tenbit = enable

        # Enabling 10bit addressing should fail (no mocked support)
        bus = SMBus(1)
        self.assertRaises(IOError, set_tenbit, bus, True)
        self.assertRaises(IOError, set_tenbit, bus, 1)
        self.assertEqual(bus.tenbit, 0)

        bus._tenbit = 1
        self.assertEqual(bus.tenbit, 1)
        bus.close()
        self.assertEqual(bus.tenbit, 0)

        # Test true enable by changing ioctl mock:
        with switch_to_full_featured_ioctl_mock():
            bus = SMBus(1)
            self.assertEqual(bus.tenbit, 0)
            set_tenbit(bus, enable=True)
            self.assertEqual(bus.tenbit, 1)
        bus.close()
        self.assertEqual(bus.tenbit, 0)


class TestSMBusWrapper(SMBusTestCase):
    """Similar test as TestSMBus except it encapsulates it all access within "with" blocks."""

    def test_func(self):
        with SMBus(1) as bus:
            print(f"Supported I2C functionality: 0x{bus.funcs:X}")
            self.assertTrue(bus.funcs & I2cFunc.I2C > 0)
            self.assertTrue(bus.funcs & I2cFunc.SMBUS_QUICK > 0)
            self.assertFalse(bus.funcs & I2cFunc.SMBUS_PEC > 0)

    def test_read(self):
        res = []
        res2 = []
        res3 = []

        # Read bytes
        with SMBus(1) as bus:
            for k in range(2):
                x = bus.read_byte_data(80, k)
                res.append(x)
        self.assertEqual(len(res), 2, msg=INCORRECT_LENGTH_MSG)

        # Read word
        with SMBus(1) as bus:
            x = bus.read_word_data(80, 0)
            res2.append(x & 255)
            res2.append(x / 256)
        self.assertEqual(len(res2), 2, msg=INCORRECT_LENGTH_MSG)
        self.assertListEqual(res, res2, msg="Byte and word reads differ")

        # Read block of N bytes
        n = 2
        with SMBus(1) as bus:
            x = bus.read_i2c_block_data(80, 0, n)
            res3.extend(x)
        self.assertEqual(len(res3), n, msg=INCORRECT_LENGTH_MSG)
        self.assertListEqual(res, res3, msg="Byte and block reads differ")


class TestI2CMsg(SMBusTestCase):
    def test_i2c_msg_rd(self):
        # 1: Convert message content to list
        msg = i2c_msg.read(60, 10)
        self.assertEqual(len(list(msg)), 10)

        # 2: Test dunder methods
        print(f"Testing dunder methods for i2c_msg: {msg.__repr__()}")
        self.assertEqual(str(msg), "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")
        self.assertEqual(
            repr(msg), "i2c_msg(60,1,b'\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00')"
        )
        self.assertEqual(bytes(msg), b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")

    def test_i2c_msg_wr(self):
        # 1: Convert message content to list
        msg = i2c_msg.write(60, [1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        data = list(msg)
        self.assertEqual(len(data), 10)

        # 2: i2c_msg is iterable
        k = 0
        s = 0
        for value in msg:
            k += 1
            s += value
        self.assertEqual(k, 10, msg="Incorrect length")
        self.assertEqual(s, 55, msg="Incorrect sum")

        # 3: Through i2c_msg properties
        s = 0
        for k in range(0, msg.len):
            s += ord(msg.buf[k])
        k += 1  # Convert to length
        self.assertEqual(k, 10, msg="Incorrect length")
        self.assertEqual(s, 55, msg="Incorrect sum")

        # 4: Test dunder methods
        print(f"Testing dunder methods for i2c_msg: {msg.__repr__()}")
        self.assertEqual(str(msg), "\x01\x02\x03\x04\x05\x06\x07\x08\t\n")
        self.assertEqual(
            repr(msg), "i2c_msg(60,0,b'\\x01\\x02\\x03\\x04\\x05\\x06\\x07\\x08\\t\\n')"
        )
        self.assertEqual(bytes(msg), b"\x01\x02\x03\x04\x05\x06\x07\x08\t\n")

    def test_i2c_msg_wr_str(self):
        # 1: Convert message content to list
        msg = i2c_msg.write(60, "foo")
        data = list(msg)
        self.assertEqual(len(data), 3)

        # 2: i2c_msg is iterable
        k = 0
        s = 0
        for value in msg:
            k += 1
            s += value
        self.assertEqual(k, 3, msg="Incorrect length")
        self.assertEqual(s, 324, msg="Incorrect sum")

        # 3: Through i2c_msg properties
        s = 0
        for k in range(0, msg.len):
            s += ord(msg.buf[k])
        k += 1  # Convert to length
        self.assertEqual(k, 3, msg="Incorrect length")
        self.assertEqual(s, 324, msg="Incorrect sum")

        # 4: Test dunder methods
        print(f"Testing dunder methods for i2c_msg: {msg.__repr__()}")
        self.assertEqual(str(msg), "foo")
        self.assertEqual(repr(msg), "i2c_msg(60,0,b'foo')")
        self.assertEqual(bytes(msg), b"foo")


class TestI2CMsgRDWR(SMBusTestCase):
    def test_i2c_rdwr_single_rd(self):
        msg = i2c_msg.read(60, 10)
        self.assertEqual(len(list(msg)), 10)
        with SMBus(1) as bus:
            bus.i2c_rdwr(msg)

    def test_i2c_rdwr_single_wr(self):
        msg = i2c_msg.write(60, "foo")
        with SMBus(1) as bus:
            bus.i2c_rdwr(msg)

    def test_i2c_rdwr_multi(self):
        msg1 = i2c_msg.read(60, 10)
        msg2 = i2c_msg.write(60, "foo")
        msg3 = i2c_msg.write(60, "bar")
        msg4 = i2c_msg.write(60, "baz")
        with SMBus(1) as bus:
            bus.i2c_rdwr(msg1, msg2, msg3, msg4)

    def test_i2c_rd(self):
        with SMBus(1) as bus:
            x = bus.i2c_rd(60, 10)
            self.assertTrue(x.flags & I2C_M_Bitflag.I2C_M_RD > 0)
            self.assertFalse(x.flags & I2C_M_Bitflag.I2C_M_TEN > 0)
            self.assertFalse(x.flags & I2C_M_Bitflag.I2C_M_WR > 0)
            # Test that we can set alternative bitflags:
            x = bus.i2c_rd(60, 10, flags=I2C_M_Bitflag.I2C_M_RD_TEN)
            self.assertTrue(x.flags & I2C_M_Bitflag.I2C_M_RD > 0)
            self.assertTrue(x.flags & I2C_M_Bitflag.I2C_M_TEN > 0)
            self.assertTrue(x.flags & I2C_M_Bitflag.I2C_M_RD_TEN > 0)
            self.assertFalse(x.flags & I2C_M_Bitflag.I2C_M_WR > 0)

    def test_i2c_wr(self):
        with SMBus(1) as bus:
            x = bus.i2c_wr(60, [1, 2, 3])
            self.assertFalse(x.flags & I2C_M_Bitflag.I2C_M_RD > 0)
            self.assertFalse(x.flags & I2C_M_Bitflag.I2C_M_TEN > 0)

            # Test that we can set alternative bitflags:
            x = bus.i2c_wr(60, [1, 2, 3], flags=I2C_M_Bitflag.I2C_M_WR_TEN)
            self.assertTrue(x.flags & I2C_M_Bitflag.I2C_M_TEN > 0)
            self.assertTrue(x.flags & I2C_M_Bitflag.I2C_M_WR_TEN > 0)
            self.assertFalse(x.flags & I2C_M_Bitflag.I2C_M_RD > 0)



from collections.abc import Generator, Iterable, Sequence
from ctypes import Array, Structure, Union, c_uint8, c_uint16, c_uint32, pointer
from enum import IntFlag
from types import TracebackType
from typing import SupportsBytes

I2C_RETRIES: int
I2C_TIMEOUT: int
I2C_SLAVE: int
I2C_SLAVE_FORCE: int
I2C_TENBIT: int
I2C_FUNCS: int
I2C_RDWR: int
I2C_PEC: int
I2C_SMBUS: int
I2C_SMBUS_WRITE: int
I2C_SMBUS_READ: int
I2C_SMBUS_QUICK: int
I2C_SMBUS_BYTE: int
I2C_SMBUS_BYTE_DATA: int
I2C_SMBUS_WORD_DATA: int
I2C_SMBUS_PROC_CALL: int
I2C_SMBUS_BLOCK_DATA: int
I2C_SMBUS_BLOCK_PROC_CALL: int
I2C_SMBUS_I2C_BLOCK_DATA: int
I2C_SMBUS_BLOCK_MAX: int

class I2C_M_Bitflag(IntFlag):
    I2C_M_RD: int = ...
    I2C_M_WR: int = ...
    I2C_M_TEN: int = ...
    I2C_M_RD_TEN: int = ...
    I2C_M_WR_TEN: int = ...

class I2cFunc(IntFlag):
    I2C: int = ...
    ADDR_10BIT: int = ...
    PROTOCOL_MANGLING: int = ...
    SMBUS_PEC: int = ...
    NOSTART: int = ...
    SLAVE: int = ...
    SMBUS_BLOCK_PROC_CALL: int = ...
    SMBUS_QUICK: int = ...
    SMBUS_READ_BYTE: int = ...
    SMBUS_WRITE_BYTE: int = ...
    SMBUS_READ_BYTE_DATA: int = ...
    SMBUS_WRITE_BYTE_DATA: int = ...
    SMBUS_READ_WORD_DATA: int = ...
    SMBUS_WRITE_WORD_DATA: int = ...
    SMBUS_PROC_CALL: int = ...
    SMBUS_READ_BLOCK_DATA: int = ...
    SMBUS_WRITE_BLOCK_DATA: int = ...
    SMBUS_READ_I2C_BLOCK: int = ...
    SMBUS_WRITE_I2C_BLOCK: int = ...
    SMBUS_HOST_NOTIFY: int = ...
    SMBUS_BYTE: int = ...
    SMBUS_BYTE_DATA: int = ...
    SMBUS_WORD_DATA: int = ...
    SMBUS_BLOCK_DATA: int = ...
    SMBUS_I2C_BLOCK: int = ...
    SMBUS_EMUL: int = ...

I2C_M_RD: int
I2C_M_WR: int
I2C_M_TEN: int
LP_c_uint8: type[pointer[c_uint8]]  # type: ignore[valid-type]
LP_c_uint16: type[pointer[c_uint16]]  # type: ignore[valid-type]
LP_c_uint32: type[pointer[c_uint32]]  # type: ignore[valid-type]

i2c_smbus_data: Array = ...

class union_i2c_smbus_data(Union): ...

union_pointer_type: pointer[union_i2c_smbus_data]  # type: ignore[valid-type]

class i2c_smbus_ioctl_data(Structure):
    @staticmethod
    def create(read_write: int = ..., command: int = 0, size: int = ...) -> i2c_smbus_ioctl_data: ...

class i2c_msg(Structure):
    def __iter__(self) -> Generator[int, None, None]: ...
    def __len__(self) -> int: ...
    def __bytes__(self) -> str: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    @staticmethod
    def read(address: int, length: int, flags: int = ...) -> i2c_msg: ...
    @staticmethod
    def write(
        address: int, buf: str | Iterable[int] | SupportsBytes, flags: int = ...
    ) -> i2c_msg: ...

class i2c_rdwr_ioctl_data(Structure):
    @staticmethod
    def create(*i2c_msg_instances: Sequence[i2c_msg]) -> i2c_rdwr_ioctl_data: ...

class SMBus:
    fd: int | None = ...
    funcs: I2cFunc = ...
    address: int | None = ...
    force: bool = ...
    pec: int = ...
    retries: int = ...
    tenbit: int = ...
    timeout: int = ...
    def __init__(self, bus: None | int | str = ..., force: bool = ...) -> None: ...
    def __enter__(self) -> SMBus: ...
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None: ...
    def open(self, bus: int | str) -> None: ...
    def close(self) -> None: ...
    def enable_pec(self, enable: bool = True) -> None: ...
    def enable_tenbit(self, enable: bool = True) -> None: ...
    def set_timeout(self, timeout: int) -> None: ...
    def set_retries(self, retries: int) -> None: ...
    def write_quick(self, i2c_addr: int, force: bool | None = None) -> None: ...
    def read_byte(self, i2c_addr: int, force: bool | None = None) -> int: ...
    def write_byte(self, i2c_addr: int, value: int, force: bool | None = None) -> None: ...
    def read_byte_data(self, i2c_addr: int, register: int, force: bool | None = None) -> int: ...
    def write_byte_data(
        self, i2c_addr: int, register: int, value: int, force: bool | None = None
    ) -> None: ...
    def read_word_data(self, i2c_addr: int, register: int, force: bool | None = None) -> int: ...
    def write_word_data(
        self, i2c_addr: int, register: int, value: int, force: bool | None = None
    ) -> None: ...
    def process_call(self, i2c_addr: int, register: int, value: int, force: bool | None = None): ...
    def read_block_data(
        self, i2c_addr: int, register: int, force: bool | None = None
    ) -> list[int]: ...
    def write_block_data(
        self,
        i2c_addr: int,
        register: int,
        data: Sequence[int],
        force: bool | None = None,
    ) -> None: ...
    def block_process_call(
        self,
        i2c_addr: int,
        register: int,
        data: Sequence[int],
        force: bool | None = None,
    ) -> list[int]: ...
    def read_i2c_block_data(
        self, i2c_addr: int, register: int, length: int, force: bool | None = None
    ) -> list[int]: ...
    def write_i2c_block_data(
        self,
        i2c_addr: int,
        register: int,
        data: Sequence[int],
        force: bool | None = None,
    ) -> None: ...
    def i2c_rdwr(self, *i2c_msgs: i2c_msg) -> None: ...
    def i2c_rd(self, i2c_addr: int, length: int, flags: int = ...) -> i2c_msg: ...
    def i2c_wr(self, i2c_addr: int, buf: Sequence[int], flags: int = ...) -> None: ...



import unittest

from smbus3.smbus3 import (  # noqa: F401
    I2C_SMBUS_BLOCK_MAX,
    I2C_SMBUS_BYTE_DATA,
    I2C_SMBUS_READ,
    i2c_msg,
    i2c_rdwr_ioctl_data,
    i2c_smbus_ioctl_data,
    union_i2c_smbus_data,
)


class TestDataTypes(unittest.TestCase):
    def test_union_i2c_smbus_data(self):
        u = union_i2c_smbus_data()

        # Fill array with values 1, 2, ...
        for k in range(I2C_SMBUS_BLOCK_MAX + 2):
            u.block[k] = k + 1

        # Check that the union works
        self.assertEqual(u.byte, u.block[0], msg="Byte field differ")
        self.assertEqual(u.block[16], 17, msg="Array field does not match")

        # Set byte and se it reflected in the array
        u.byte = 255
        self.assertEqual(u.block[0], 255, msg="Byte field not reflected in array")

        # Reset array to zeros and check word field
        for k in range(I2C_SMBUS_BLOCK_MAX + 2):
            u.block[k] = 0
        u.word = 1607
        self.assertNotEqual(0, u.word, msg="Word field is zero but should be non-zero")
        u.word = 0

    def test_i2c_smbus_ioctl_data_factory(self):
        ioctl_msg = i2c_smbus_ioctl_data.create()

        self.assertEqual(ioctl_msg.read_write, I2C_SMBUS_READ)
        self.assertEqual(ioctl_msg.size, I2C_SMBUS_BYTE_DATA)

        # Simple test to check assignment
        ioctl_msg.data.contents.byte = 25
        self.assertEqual(ioctl_msg.data.contents.byte, 25, msg="Get not equal to set")

    def test_i2c_msg_read(self):
        msg = i2c_msg.read(80, 10)
        self.assertEqual(msg.addr, 80)
        self.assertEqual(msg.len, 10)
        self.assertEqual(msg.len, len(msg))
        self.assertEqual(msg.flags, 1)

    def test_i2c_msg_write(self):
        # Create from list
        buf = [65, 66, 67, 68, 1, 10, 255]
        msg = i2c_msg.write(81, buf)
        self.assertEqual(msg.addr, 81)
        self.assertEqual(msg.len, 7)
        self.assertEqual(msg.len, len(msg))
        self.assertEqual(msg.flags, 0)
        self.assertListEqual(buf, list(msg))
        # Create from str
        s = "ABCD\x01\n\xff"
        msg2 = i2c_msg.write(81, s)
        self.assertEqual(msg2.addr, msg.addr)
        self.assertEqual(msg2.len, msg.len)
        self.assertEqual(msg2.flags, msg.flags)
        self.assertListEqual(list(msg), list(msg2))
        self.assertEqual(str(msg2), "ABCD\x01\n")
        self.assertGreaterEqual(f"{msg2!r}".find(r"ABCD\x01\n\xff"), 0)

    def test_i2c_msg_iter(self):
        buf = [10, 11, 12, 13]
        msg = i2c_msg.write(81, buf)
        # Convert msg to list and compare
        msg_list = list(msg)
        self.assertListEqual(buf, msg_list)
        # Loop over each message entry
        i = 0
        for value in msg:
            self.assertEqual(value, buf[i])
            i += 1
        # Loop over with index and value
        for i, value in enumerate(msg):
            self.assertEqual(i + 10, value)
